import argparse
import math
import os
import pickle
import time
import matplotlib
import sys

import pandas as pd
import torch
import numpy as np
from torch.hub import download_url_to_file
import matplotlib.pyplot as plt
import torch.utils.data
import torch.nn.functional as F
from tqdm import tqdm
import json

from torch.nn.utils.rnn import pack_padded_sequence, PackedSequence, pad_packed_sequence

plt.rcParams["figure.figsize"] = (10, 10) # size of window
plt.style.use('dark_background')

parser = argparse.ArgumentParser()

parser.add_argument('-device', type=str, default='cpu')
parser.add_argument('-batch_size', type=int, default=16)
parser.add_argument('-learning_rate', type=float, default=1e-3)

parser.add_argument('-train_test_split', type=float, default=0.8)

parser.add_argument('-input_captcha_width', type=int, default=128)
parser.add_argument('-input_captcha_height', type=int, default=64)
parser.add_argument('-input_captcha_length', type=int, default=4)

parser.add_argument('-model_embedding_size', type=int, default=4)
parser.add_argument('-model_hidden_size', type=int, default=4)

parser.add_argument('-num_epochs', type=int, default=10)

args = parser.parse_args()

MAX_LEN = 200
if torch.cuda.is_available():
    args.device = 'cuda'
    # comment out this next line if you have nvidia GPU and you want to debug
    MAX_LEN = None


class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        path_dataset = '../data/cs448b_ipasn.csv'
        if not os.path.exists(path_dataset):
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                'http://share.yellowrobot.xyz/1658433567-dl-summer-school/cs448b_ipasn.csv',
                path_dataset,
                progress=True
            )

        df_full = pd.read_csv('../data/cs448b_ipasn.csv')
        df_clean = df_full.dropna()
        df_grouped = df_clean.groupby(["date", "l_ipn"])
        date_by_l_ipn = list(df_grouped['date'].count().index)

        self.max_seq_length = 0

        self.isp_source_ids = []
        self.fs = []


        sequences_raw = []
        for date, l_ipn in date_by_l_ipn:
            sequence = df_clean[(df_clean['date'] == date) & (df_clean['l_ipn'] == l_ipn)][["r_asn", "f"]].values
            for r_asn, f in sequence:
                if r_asn not in self.isp_source_ids:
                    self.isp_source_ids.append(r_asn)
                self.fs.append(f)
            self.max_seq_length = max(self.max_seq_length, len(sequence))
            sequences_raw.append(sequence)

        # replace ISPs with indexes
        # standardize the call count
        self.sequences = []
        self.f_mean = np.mean(self.fs)
        self.f_std = np.std(self.fs)
        for sequence in sequences_raw:
            sequence_standardized = []
            for r_asn, f in sequence:
                sequence_standardized.append(
                    (
                        self.isp_source_ids.index(r_asn),
                        (f - self.f_mean) / self.f_std
                    ))
            self.sequences.append(sequence_standardized)

        # 847 samples
        # 96 max sequence length
        # 2 input features

        # group by date and l_ipn (local ip address)
        # maintain order and check r_asn (remote ISP), f (flows per day)

        # task detect suspicious activity at these dates (should get much higher accuracy for a model if you remove these dates)
        # Date : IP
        # 08-24 : 1
        # 09-04 : 5
        # 09-18 : 4
        # 09-26 : 3 6

    def __len__(self):
        if MAX_LEN:
            return MAX_LEN
        return len(self.sequences)

    def __getitem__(self, idx):
        x_raw = self.sequences[idx]

        y = np.roll(x_raw, -1) # move all items by one position left
        y = y[:-1]
        x = x_raw[:-1]
        x_length = len(x)

        x_padded = np.zeros((self.max_seq_length, 2), dtype=np.float32)
        y_padded = np.zeros((self.max_seq_length, 2), dtype=np.float32)

        x_padded[:x_length] = x
        y_padded[:x_length] = y

        return x_padded, y_padded, x_length

dataset_full = Dataset()

train_test_split = int(len(dataset_full) * args.train_test_split)
dataset_train, dataset_test = torch.utils.data.random_split(
    dataset_full,
    [train_test_split, len(dataset_full) - train_test_split],
    generator=torch.Generator().manual_seed(0)
)

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=args.batch_size,
    shuffle=True,
    drop_last=(len(dataset_train) % args.batch_size < args.batch_size * 0.5)
)

dataloader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=args.batch_size
)


class RNN(torch.nn.Module):
    def __init__(self, input_size, hidden_size):
        super().__init__()

        self.hidden_size = hidden_size
        # TODO implement

    def forward(self, x: PackedSequence, hidden=None):
        # TODO implement
        out_packed = x
        return out_packed, hidden


class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.isp_embedding = torch.nn.Embedding(
            num_embeddings=len(dataset_full.isp_source_ids),
            embedding_dim=args.model_embedding_size
        )

        self.rnn = RNN(
            args.model_embedding_size + 1,
            args.model_hidden_size
        )

        self.fc = torch.nn.Linear(
            in_features=args.model_hidden_size,
            out_features=len(dataset_full.isp_source_ids) + 1 # probability of each ISP + 1 f for the flow
        )

    def forward(self, x: PackedSequence, hidden=None):

        x_emb = self.isp_embedding.forward(x.data[:, :1].long())
        x_zipped = torch.cat([x_emb.squeeze(), x.data[:, 1:]], dim=-1)

        x_emb_seq = PackedSequence(
            data=x_zipped,
            batch_sizes=x.batch_sizes,
            sorted_indices=x.sorted_indices
        )

        out, hidden = self.rnn.forward(x_emb_seq, hidden)

        out_fc = self.fc.forward(out.data)  # out.data

        y_prim_packed = PackedSequence(
            data=out_fc,
            batch_sizes=x.batch_sizes,
            sorted_indices=x.sorted_indices
        )

        return y_prim_packed, hidden


model = Model()
model = model.to(args.device)

optimizer = torch.optim.Adam(
    model.parameters(),
    lr=args.learning_rate
)

loss_plot_train = []
loss_plot_test = []

for epoch in range(1, 1000):

    for dataloader in [dataloader_train, dataloader_test]:
        losses = []
        r2s = []
        for x_padded, y_padded, x_length in tqdm(dataloader):

            x_padded = x_padded.to(args.device)
            y_padded = y_padded.to(args.device)

            x_packed = pack_padded_sequence(x_padded, x_length, batch_first=True, enforce_sorted=False)
            y_packed = pack_padded_sequence(y_padded, x_length, batch_first=True, enforce_sorted=False)

            y_prim_packed, _ = model.forward(x_packed)
            loss_reg = torch.mean(torch.abs(y_packed.data[:, 1] - y_prim_packed.data[:, 1]))
            loss_cce = torch.nn.functional.cross_entropy(y_prim_packed.data[:, :1], y_packed.data[:, :1])

            loss = loss_cce + loss_reg
            losses.append(loss.cpu().item())

            if dataloader == dataloader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

        if dataloader == dataloader_train:
            loss_plot_train.append(np.mean(losses))
        else:
            loss_plot_test.append(np.mean(losses))

    print(
        f'\n\nepoch: {epoch} '
        f'loss_train: {loss_plot_train[-1]} '
        f'loss_test: {loss_plot_test[-1]} '
    )

    if epoch % 20 == 0 or True:
        plt.clf()
        plt.plot(loss_plot_train, 'r-', label='loss train')
        plt.legend()
        plt.xlabel("Epoch")
        ax2 = plt.twinx()
        ax2.plot(loss_plot_test, 'c-', label='loss test')
        ax2.legend(loc='upper left')

        plt.tight_layout(pad=0.5)
        plt.show()


